from tensorflow_modules import *
import tensorflow as tf
import unittest


class TestAttention(unittest.TestCase):
    batch_size = 10
    num_heads = 8
    depth = 512

    def test_dimension(self):
        q = tf.random.uniform([self.batch_size, self.num_heads, 1, self.depth])
        k = tf.random.uniform([self.batch_size, self.num_heads, 2, self.depth])
        v = tf.random.uniform([self.batch_size, self.num_heads, 2, self.depth])

        output, attention_weights = scaled_dot_product_attention(q, k, v)
        self.assertEqual([self.batch_size, self.num_heads, 1, self.depth], output.shape)
        self.assertEqual([self.batch_size, self.num_heads, 1, 2], attention_weights.shape)

    def test_padding_mask(self):
        q = tf.random.uniform([self.batch_size, self.num_heads, 1, self.depth])
        k = tf.random.uniform([self.batch_size, self.num_heads, 2, self.depth])
        v = tf.random.uniform([self.batch_size, self.num_heads, 2, self.depth])

        padding_mask = tf.random.uniform([self.batch_size, 1, 1, 2])

        output, attention_weights = scaled_dot_product_attention(q, k, v, padding_mask)
        self.assertEqual([self.batch_size, self.num_heads, 1, self.depth], output.shape)
        self.assertEqual([self.batch_size, self.num_heads, 1, 2], attention_weights.shape)

    def test_look_ahead_mask(self):
        q = tf.random.uniform([self.batch_size, self.num_heads, 1, self.depth])
        k = tf.random.uniform([self.batch_size, self.num_heads, 2, self.depth])
        v = tf.random.uniform([self.batch_size, self.num_heads, 2, self.depth])

        look_ahead_mask = tf.random.uniform([1, 1])

        output, attention_weights = scaled_dot_product_attention(q, k, v, look_ahead_mask)
        self.assertEqual([self.batch_size, self.num_heads, 1, self.depth], output.shape)
        self.assertEqual([self.batch_size, self.num_heads, 1, 2], attention_weights.shape)



from tensorflow_modules import *
import tensorflow as tf
import unittest


class TestEncoderLayer(unittest.TestCase):
    d_model = 512
    num_heads = 8
    dff = 2048

    def test_dimension(self):
        model = EncoderLayer(self.d_model, self.num_heads, self.dff)
        x = tf.random.uniform([64, 43, 512])

        output = model(x, False, None)
        self.assertEqual([64, 43, 512], output.shape.as_list())



import tensorflow as tf


def scaled_dot_product_attention(q, k, v, mask=None):
    """
    アテンションの重みの計算
    q, k, vは最初の次元が一致していること
    k, vは最後から2番めの次元が一致していること
    マスクは型（パディングかルックアヘッドか）によって異なるshapeを持つが、
    加算の際にブロードキャスト可能であること
    引数：
        q: query shape == (..., seq_len_q, depth)
        k: key shape == (..., seq_len_k, depth)
        v: value shape == (..., seq_len_v, depth_v)
        mask: (..., seq_len_q, seq_len_k) にブロードキャスト可能な
                    shapeを持つ浮動小数点テンソル。既定値はNone

    戻り値：
        出力、アテンションの重み
    """

    matmul_qk = tf.matmul(q, k, transpose_b=True)   # (..., seq_len_q, seq_len_k)

    # matmul_qkをスケール
    dk = tf.cast(tf.shape(k)[-1], tf.float32)
    scaled_attention_logits = matmul_qk / tf.math.sqrt(dk)

    # マスクをスケール済みテンソルに加算
    if mask is not None:
        scaled_attention_logits += (mask * -1e9)

    # softmax は最後の軸(seq_len_k)について
    # 合計が1となるように正規化
    attention_weights = tf.nn.softmax(scaled_attention_logits, axis=-1)     # (..., seq_len_q, seq_len_k)

    output = tf.matmul(attention_weights, v)    # (..., seq_len_q, depth_v)

    return output, attention_weights



from .decoder_layer import DecoderLayer
from .encoder_layer import EncoderLayer
from .point_wise_feed_forward_network import PointWiseFeedForwardNetwork


__all__ = ['DecoderLayer', 'EncoderLayer', 'PointWiseFeedForwardNetwork']



from .reversible_decoder_layer import ReversibleDecoderLayer
from .reversible_decoder_layer import ReversalReversibleDecoderLayer
from .reversible_encoder_layer import ReversibleEncoderLayer


__all__ = ['ReversibleDecoderLayer', 'ReversibleEncoderLayer', 'ReversalReversibleDecoderLayer']



from ..transformer.point_wise_feed_forward_network import PointWiseFeedForwardNetwork
from ..multi_head_attention.model import MultiHeadAttention
import tensorflow as tf


class ReversibleEncoderLayer(tf.keras.layers.Layer):
    def __init__(self, d_model, num_heads, dff, rate=0.1):
        super().__init__()

        self.mha = MultiHeadAttention(d_model, num_heads)
        self.ffn = PointWiseFeedForwardNetwork(d_model, dff)

        self.layernorm1 = tf.keras.layers.LayerNormalization(epsilon=1e-6)
        self.layernorm2 = tf.keras.layers.LayerNormalization(epsilon=1e-6)

        self.dropout1 = tf.keras.layers.Dropout(rate)
        self.dropout2 = tf.keras.layers.Dropout(rate)

    def call(self, left, right, training, mask):
        attn_output, _ = self.mha(left, left, left, mask)   # (batch_size, input_seq_len, d_model)
        attn_output = self.dropout1(attn_output, training=training)
        right = self.layernorm1(right + attn_output)        # (batch_size, input_seq_len, d_model)

        ffn_output = self.ffn(right)                # (batch_size, input_seq_len, d_model)
        ffn_output = self.dropout2(ffn_output, training=training)
        left = self.layernorm2(left + ffn_output)   # (batch_size, input_seq_len, d_model)

        return left, right



from tensorflow_modules import *
import tensorflow as tf
import unittest


class TestReversibleDecoderLayer(unittest.TestCase):
    d_model = 512
    num_heads = 8
    dff = 2048

    def test_dimension(self):
        model = ReversibleDecoderLayer(self.d_model, self.num_heads, self.dff)
        left = tf.random.uniform([64, 43, 512])
        right = tf.random.uniform([64, 43, 512])
        enc_output = tf.random.uniform([64, 43, 512])

        left_out, right_out, attn1, attn2 = model(left, right, enc_output, False, None, None)
        self.assertEqual([64, 43, 512], left_out.shape.as_list())
        self.assertEqual([64, 43, 512], right_out.shape.as_list())
        self.assertEqual([64, 8, 43, 43], attn1.shape)
        self.assertEqual([64, 8, 43, 43], attn2.shape)


class TestReversalReversibleDecoderLayer(unittest.TestCase):
    d_model = 512
    num_heads = 8
    dff = 2048

    def test_dimension(self):
        model = ReversalReversibleDecoderLayer(self.d_model, self.num_heads, self.dff)
        left = tf.random.uniform([64, 43, 512])
        right = tf.random.uniform([64, 43, 512])
        enc_output = tf.random.uniform([64, 43, 512])

        left_out, right_out, attn1, attn2 = model(left, right, enc_output, False, None, None)
        self.assertEqual([64, 43, 512], left_out.shape.as_list())
        self.assertEqual([64, 43, 512], right_out.shape.as_list())
        self.assertEqual([64, 8, 43, 43], attn1.shape)
        self.assertEqual([64, 8, 43, 43], attn2.shape)



import tensorflow as tf


class PointWiseFeedForwardNetwork(tf.keras.layers.Layer):
    def __init__(self, d_model, dff):
        super(PointWiseFeedForwardNetwork, self).__init__()

        self.ffn = tf.keras.Sequential([
            tf.keras.layers.Dense(dff, activation='relu'),  # (batch_size, seq_len, dff)
            tf.keras.layers.Dense(d_model)                  # (batch_size, seq_len, d_model)
        ])

    def call(self, x):
        return self.ffn(x)
